@@ -5,7 +5,7 @@ from contextlib import contextmanager, redirect_stdout
 from copy import copy, deepcopy
 from functools import wraps, partial, update_wrapper
 import inspect
-from inspect import Parameter, signature
+from inspect import Parameter, signature, _empty
 import io
 import json
 import logging
@@ -2495,3 +2495,135 @@ def get_module_docstring(path, default=''):
         tree = ast.parse(f.read())
     return ast.get_docstring(tree) or default
 
+
+def getindex(arr, val, default=-1):
+    """Like list.index but allows us to specify a fallback value if the value
+    isn't present in the list, kind of like getattr.
+
+    Parameters
+    ----------
+    arr: list
+        The list to search in.
+    val: any
+        The item to search for.
+    default: any
+        Value to return if val is not in arr.
+    """
+    return arr.index(val) if val in arr else default
+
+
+class Partial:
+    """More powerful (though also potentially more fragile) version of
+    functools.partial that updates the resulting signature to work better with
+    Jupyter's quick documentation feature. We also update __repr__, __str__,
+    and __name__ attributes (optionally renaming the source function). Unlike
+    functools.partial, we also reorder parameters so that those without
+    defaults always come before those with defaults.
+
+    Note: the resulting object is actually a callable class, not a function.
+    """
+
+    def __init__(self, func, name=None, **kwargs):
+        """
+        Parameters
+        ----------
+        func: function
+        name: str or None
+            If None, the source function's name will be used.
+        kwargs: any
+            Default arguments to set, like in functools.partial.
+        """
+        self.func = copy_func(func)
+        self.old_name = func.__name__
+
+        # Track names of positional args in old function since this affects
+        # the order args must be passed in if var_positional parameters
+        # (*args) are present.
+        self.old_pos_pars = []
+        self.kwargs_name = ''
+        self.args_name = ''
+        new_pars = []
+        old_sig = signature(self.func)
+        for k, v in old_sig.parameters.items():
+            # Check parameter kind for error handling and argument resolution
+            # in __call__.
+            if v.kind == 0:
+                raise NotImplementedError(
+                    'rigorous_partial does not support functions with '
+                    'positional only parameters.'
+                )
+            elif v.kind == 2:
+                self.args_name = k
+            elif v.kind == 4:
+                self.kwargs_name = k
+                break
+
+            if v.kind <= 2:
+                self.old_pos_pars.append(k)
+
+            # Assign default value from newly specified kwargs if provided.
+            if k in kwargs:
+                default = kwargs.pop(k)
+                kind = 3
+            else:
+                default = v.default
+                kind = v.kind
+            param = Parameter(k, kind, default=default)
+            new_pars.append(param)
+
+        # Remaining kwargs only: those that were not present in func's
+        # signature. Require that they be keyword only since ordering can
+        # cause issues (updating signature affects what we see but doesn't
+        # seem to affect the actual order args are passed in, presumably due
+        # to old __code__ object).
+        for k, v in kwargs.items():
+            param = Parameter(k, 3, default=v)
+            new_pars.append(param)
+        if self.kwargs_name:
+            new_pars.append(Parameter(self.kwargs_name, 4))
+
+        # Ensure we don't accidentally place any parameters with defaults
+        # ahead of those without them. Third item in tuple is a tiebreaker
+        # (defaults to original function's parameter order).
+        old_names = [p for p in old_sig.parameters]
+        new_pars.sort(
+            key=lambda x: (x.kind, x.default != _empty,
+                           getindex(old_names, x.name, float('inf')))
+        )
+
+        # I honestly forget why we need to set the attribute on self.func too,
+        # I just remember it was needed to resolve a bug (I think it was
+        # related to *args resolution).
+        self.__signature__ = self.func.__signature__ = old_sig.replace(
+            parameters=new_pars
+        )
+        self.__defaults__ = tuple(p.default for p in new_pars if p.kind < 3
+                                  and p.default != _empty)
+        self.__kwdefaults__ = {p.name: p.default for p in new_pars
+                               if p.kind == 3}
+        if name: self.func.__name__ = name
+        update_wrapper(self, self.func)
+
+    def __call__(self, *args, **new_kwargs):
+        # Remember self.func's actual code is unchanged: we updated how its
+        # signature appears, but that doesn't affect the actual mechanics.
+        # Therefore, we need to carefully resolve args and kwargs so that the
+        # function is called so that behavior matches what we'd expect based
+        # on the order shown in the signature.
+        tmp_kwargs = bound_args(self.func, args,
+                                {**self.__kwdefaults__, **new_kwargs})
+        final_args = {name: tmp_kwargs.pop(name)
+                      for name in self.old_pos_pars}
+        final_star_args = final_args.pop(self.args_name, [])
+        final_kwargs = select(tmp_kwargs, drop=list(final_args))
+        return self.func(*final_args.values(), *final_star_args,
+                         **final_kwargs)
+
+    def __repr__(self):
+        """Note: the memory address here points to that of the copy of the
+        source function stored in self.func.
+        """
+        return repr(self.func).replace(self.old_name, self.__name__)
+
+    def __str__(self):
+        return str(self.func).replace(self.old_name, self.__name__)
